@@ -13,6 +13,8 @@
 
 #include "column.h"
 #include "macros.h"
+#include "timing.h"
+
 #if HAS_INCLUDE(<pthread_np.h>)
 #include <pthread_np.h>
 #define pthread_setname_np(thread, name) (pthread_set_name_np(thread, name), 0)
@@ -32,6 +34,10 @@
  #include <nanosvg/nanosvgrast.h>
 #endif
 
+#if defined(FUZZEL_ENABLE_SVG_RESVG)
+ #include <resvg.h>
+#endif
+
 #define LOG_MODULE "render"
 #define LOG_ENABLE_DBG 0
 #include "log.h"
@@ -59,6 +65,7 @@ struct render {
 
     /* Cached fcft text runs */
     struct fcft_text_run *prompt_text_run;
+    struct fcft_text_run *message_text_run;
     struct fcft_text_run *placeholder_text_run;
 
     /* Cached selection corners */
@@ -72,6 +79,7 @@ struct render {
     pixman_color_t pix_background_color;
     pixman_color_t pix_border_color;
     pixman_color_t pix_text_color;
+    pixman_color_t pix_message_color;
     pixman_color_t pix_prompt_color;
     pixman_color_t pix_input_color;
     pixman_color_t pix_match_color;
@@ -89,6 +97,7 @@ struct render {
     unsigned selection_border_radius;
     unsigned row_height;
     unsigned icon_height;
+    unsigned message_height;
 
     unsigned input_glyph_offset; /* At which glyph to start rendering input */
 
@@ -544,6 +553,173 @@ adjust_input_glyph_offset(struct render *render, const struct prompt *prompt,
     }
 }
 
+static void
+current_line_boundaries(char32_t* remaining_message, const struct config *conf,
+                        char32_t** current_line_end, char32_t** next_line_start)
+{
+    char32_t* next_linebreak = c32chr(remaining_message, U'\n');
+    // Unless word wrapping, the current line boundaries are the whole line
+    // If the line is shorter than the configured width, we can also immediately return
+    if (conf->message_mode != MESSAGE_MODE_WRAP
+        || (next_linebreak != NULL && next_linebreak - remaining_message <= conf->chars)
+        || (next_linebreak == NULL && c32len(remaining_message) <= conf->chars)
+    ) {
+        if (next_linebreak) {
+            *current_line_end = next_linebreak-1;
+            // Do not skip leading spaces after a linebreak,
+            // as intentional spaces likely exist for formatting purposes.
+            *next_line_start = next_linebreak+1;
+        } else {
+            *current_line_end = remaining_message + c32len(remaining_message) - 1;
+            *next_line_start = NULL;
+        }
+        goto strip_trailing_spaces;
+    }
+
+    // Wrapping logic
+    char32_t* last_space = NULL;
+    *current_line_end = remaining_message;
+    while (*current_line_end - remaining_message < conf->chars) {
+        if (**current_line_end == U' ') {
+            last_space = *current_line_end;
+        }
+        (*current_line_end)++;
+    }
+
+    // If there is no space in the line, all we can really do (without
+    // changing the string) is stop at the end character and move to
+    // the next line.
+    // Might be nice to have a configurable "wrapping character" option
+    // later.
+    // current_line_end is already conf->chars characters after remaining_message
+    // here due to the space-searching loop
+    if (last_space == NULL) {
+        // Bumped one beyond conf->chars before the loop ends
+        (*current_line_end)--;
+        *next_line_start = *current_line_end + 1;
+        // Consistency. Should never happen here, but good to check.
+        goto strip_leading_spaces;
+    }
+
+
+    // If there is a space in the line, break on it.
+    *current_line_end = last_space-1;
+    *next_line_start = last_space+1;
+
+strip_leading_spaces:
+    while (*next_line_start != NULL && c32len(*next_line_start) > 0 && **next_line_start == U' ')
+        (*next_line_start)++;
+    if (c32len(*next_line_start) == 0)
+        *next_line_start = NULL;
+
+strip_trailing_spaces:
+    while (
+        *current_line_end != NULL
+        && *current_line_end > remaining_message
+        && **current_line_end == U' ')
+    {
+        (*current_line_end)--;
+    }
+
+}
+
+void
+render_message(struct render *render, struct buffer *buf)
+{
+    if (!render->conf->message) return;
+
+    struct fcft_font *font = render->font;
+    assert(font != NULL);
+
+    const char32_t *message = render->conf->message;
+    size_t message_len = c32len(message);
+
+    const enum fcft_subpixel subpixel =
+        (render->conf->colors.background.a == 1. &&
+         render->conf->colors.message.a == 1.)
+        ? render->subpixel : FCFT_SUBPIXEL_NONE;
+
+
+    int x = render->border_size + render->x_margin;
+    int y = render->border_size + render->y_margin + render_baseline(render);
+
+    /* Erase background */
+    pixman_color_t bg = render->pix_background_color;
+    //pixman_color_t bg = (pixman_color_t){0xffff, 0, 0, 0xffff};
+    pixman_image_fill_rectangles(
+        PIXMAN_OP_SRC, buf->pix[0], &bg, 1,
+        &(pixman_rectangle16_t){
+            render->border_size + render->x_margin - render->x_margin / 3,
+            render->border_size + render->y_margin,
+            buf->width - 2 * (render->border_size + render->x_margin - render->x_margin / 3),
+            render->message_height});
+
+    struct fcft_text_run *message_run = render->message_text_run;
+
+    if (fcft_capabilities() & FCFT_CAPABILITY_TEXT_RUN_SHAPING && message_run == NULL) {
+        message_run = fcft_rasterize_text_run_utf32(
+            font, message_len, message, subpixel);
+    }
+
+    char32_t* current_pos = render->conf->message;
+    char32_t* current_line_end = NULL;
+    char32_t* next_line_start = NULL;
+    while (current_pos != NULL) {
+        current_line_boundaries(current_pos, render->conf,
+                                &current_line_end, &next_line_start);
+
+        if (message_run != NULL) {
+            for (; current_pos <= current_line_end; current_pos++) {
+                int i = current_pos - render->conf->message;
+                const struct fcft_glyph *glyph = message_run->glyphs[i];
+
+                render_glyph(buf->pix[0], glyph, x, y, &render->pix_message_color);
+                x += glyph->advance.x;
+            }
+        } else {
+            char32_t prev = 0;
+            for (; current_pos <= current_line_end; current_pos++) {
+                int i = current_pos - render->conf->message;
+                const char32_t wc = message[i];
+                if (wc == U'\n') {
+                    x = render->border_size + render->x_margin;
+                    y += render->row_height;
+                    continue;
+                }
+
+                const struct fcft_glyph *glyph = fcft_rasterize_char_utf32(font, wc, subpixel);
+
+                if (glyph == NULL) {
+                    prev = wc;
+                    continue;
+                }
+
+                long x_kern;
+                fcft_kerning(font, prev, wc, &x_kern, NULL);
+
+                x += x_kern;
+
+                render_glyph(
+                    buf->pix[0], glyph, x, y, &render->pix_message_color);
+                x += glyph->advance.x;
+                x += pt_or_px_as_pixels(render, &render->conf->letter_spacing);
+
+                prev = wc;
+            }
+        }
+
+        current_pos = next_line_start;
+        x = render->border_size + render->x_margin;
+        y += render->row_height;
+    }
+
+
+    if (render->message_text_run == NULL)
+        render->message_text_run = message_run;
+    else if (message_run != render->message_text_run)
+        fcft_text_run_destroy(message_run);
+}
+
 void
 render_prompt(struct render *render, struct buffer *buf,
               const struct prompt *prompt, const struct matches *matches)
@@ -588,7 +764,7 @@ render_prompt(struct render *render, struct buffer *buf,
     }
 
     int x = render->border_size + render->x_margin;
-    int y = render->border_size + render->y_margin + render_baseline(render);
+    int y = render->border_size + render->y_margin + render_baseline(render) + render->message_height;
 
     /* Erase background */
     pixman_color_t bg = render->pix_background_color;
@@ -597,7 +773,7 @@ render_prompt(struct render *render, struct buffer *buf,
         PIXMAN_OP_SRC, buf->pix[0], &bg, 1,
         &(pixman_rectangle16_t){
             render->border_size + render->x_margin - render->x_margin / 3,
-            render->border_size + render->y_margin,
+            render->border_size + render->y_margin + render->message_height,
             buf->width - 2 * (render->border_size + render->x_margin - render->x_margin / 3),
             render->row_height});
 
@@ -956,30 +1132,28 @@ render_svg_nanosvg(struct icon *icon, int x, int y, int size,
         if (rast == NULL)
             return;
 
-        /* For gamma-correct blending */
-        uint8_t *data_16bit = NULL;
-        pixman_image_t *img_16bit = NULL;
-        uint64_t *abgr16 = NULL;
-
         float scale = svg->width > svg->height ? size / svg->width : size / svg->height;
 
         const int width = roundf(svg->width * scale);
         const int height = roundf(svg->height * scale);
 
         uint8_t *data_8bit = xmalloc(width * height * 4);
+        uint8_t *data_16bit = NULL;
+        uint64_t *abgr16 = NULL;
 
         nsvgRasterize(rast, svg, 0, 0, scale, data_8bit, width, height, width * 4);
 
-        img = pixman_image_create_bits_no_clear(
-            PIXMAN_a8b8g8r8, width, height, (uint32_t *)data_8bit, width * 4);
-
         if (gamma_correct) {
+            /* For gamma-correct blending, create 16-bit buffer and image */
             data_16bit = xmalloc(width * height * 8);
             abgr16 = (uint64_t *)data_16bit;
 
-            img_16bit = pixman_image_create_bits_no_clear(
+            img = pixman_image_create_bits_no_clear(
                 PIXMAN_a16b16g16r16, width, height, (uint32_t *)data_16bit,
                 width * 8);
+        } else {
+            img = pixman_image_create_bits_no_clear(
+                PIXMAN_a8b8g8r8, width, height, (uint32_t *)data_8bit, width * 4);
         }
 
         /* Nanosvg produces non-premultiplied ABGR, while pixman expects
@@ -1031,10 +1205,8 @@ render_svg_nanosvg(struct icon *icon, int x, int y, int size,
         }
 
         if (gamma_correct) {
-            /* Replace default image buffer with 16-bit linear buffer */
-            pixman_image_unref(img);
+            /* Free the 8-bit buffer as we've converted everything to 16-bit */
             free(data_8bit);
-            img = img_16bit;
         }
 
         nsvgDeleteRasterizer(rast);
@@ -1057,6 +1229,117 @@ render_svg_nanosvg(struct icon *icon, int x, int y, int size,
 }
 #endif /* FUZZEL_ENABLE_SVG_NANOSVG */
 
+#if defined(FUZZEL_ENABLE_SVG_RESVG)
+static void
+render_svg_resvg(struct icon *icon, int x, int y, int size,
+                 pixman_image_t *pix, cairo_t *cairo, bool gamma_correct)
+{
+#if defined(FUZZEL_ENABLE_CAIRO)
+    cairo_surface_flush(cairo_get_target(cairo));
+#endif
+
+    pixman_image_t *img = NULL;
+
+    /* Look for a cached image, at the correct size */
+    tll_foreach(icon->rasterized, it) {
+        if (it->item.size == size) {
+            img = it->item.pix;
+            break;
+        }
+    }
+
+    if (img == NULL) {
+        resvg_render_tree *tree = icon->svg;
+        resvg_size svg_size = resvg_get_image_size(tree);
+
+        if (svg_size.width == 0 || svg_size.height == 0)
+            return;
+
+        float scale = svg_size.width > svg_size.height ? size / svg_size.width : size / svg_size.height;
+
+        const int width = roundf(svg_size.width * scale);
+        const int height = roundf(svg_size.height * scale);
+
+        uint8_t *data_8bit = xmalloc(width * height * 4);
+        memset(data_8bit, 0, width * height * 4);
+
+        /* resvg renders to RGBA8888 premultiplied */
+        resvg_transform transform = resvg_transform_identity();
+        transform.a = scale;
+        transform.d = scale;
+        resvg_render(tree, transform, width, height, (char *)data_8bit);
+
+        if (gamma_correct) {
+            /* For gamma-correct blending, create 16-bit buffer and image */
+            uint8_t *data_16bit = xmalloc(width * height * 8);
+            uint64_t *abgr16 = (uint64_t *)data_16bit;
+
+            /* resvg produces premultiplied RGBA, need to convert to ABGR16 for pixman */
+            for (uint32_t *rgba = (uint32_t *)data_8bit;
+                 rgba < (uint32_t *)(data_8bit + width * height * 4);
+                 rgba++)
+            {
+                uint16_t red = (*rgba >> 0) & 0xff;
+                uint16_t green = (*rgba >> 8) & 0xff;
+                uint16_t blue = (*rgba >> 16) & 0xff;
+                uint16_t alpha = (*rgba >> 24) & 0xff;
+
+                /* Decode sRGB to linear for gamma-correct blending */
+                if (alpha == 0x00) {
+                    blue = green = red = 0x00;
+                } else {
+                    /* First, un-premultiply */
+                    blue = (uint16_t)(blue * 0xff / alpha);
+                    green = (uint16_t)(green * 0xff / alpha);
+                    red = (uint16_t)(red * 0xff / alpha);
+
+                    /* Convert to linear 16-bit */
+                    alpha |= alpha << 8;  /* Alpha already linear, expand to 16-bit */
+                    blue = srgb_decode_8_to_16(blue);
+                    green = srgb_decode_8_to_16(green);
+                    red = srgb_decode_8_to_16(red);
+
+                    /* Re-premultiply in 16-bit linear space */
+                    blue = blue * alpha / 0xffff;
+                    green = green * alpha / 0xffff;
+                    red = red * alpha / 0xffff;
+                }
+
+                *abgr16 = (uint64_t)alpha << 48 | (uint64_t)blue << 32 | (uint64_t)green << 16 | (uint64_t)red;
+                abgr16++;
+            }
+
+            /* Free the 8-bit buffer as we've converted everything to 16-bit */
+            free(data_8bit);
+
+            /* Create 16-bit pixman image */
+            img = pixman_image_create_bits_no_clear(
+                PIXMAN_a16b16g16r16, width, height, (uint32_t *)data_16bit,
+                width * 8);
+        } else {
+            /* Create 8-bit pixman image */
+            img = pixman_image_create_bits_no_clear(
+                PIXMAN_a8b8g8r8, width, height, (uint32_t *)data_8bit, width * 4);
+        }
+
+        tll_push_back(icon->rasterized, ((struct rasterized){img, size}));
+    }
+
+    const int w = pixman_image_get_width(img);
+    const int h = pixman_image_get_height(img);
+
+    pixman_image_composite32(
+        PIXMAN_OP_OVER, img, NULL, pix, 0, 0, 0, 0,
+        x + (size - w) / 2,
+        y + (size - h) / 2,
+        w, h);
+
+#if defined(FUZZEL_ENABLE_CAIRO)
+    cairo_surface_mark_dirty(cairo_get_target(cairo));
+#endif
+}
+#endif /* FUZZEL_ENABLE_SVG_RESVG */
+
 static void
 render_svg(struct icon *icon, int x, int y, int size,
            pixman_image_t *pix, cairo_t *cairo, bool gamma_correct,
@@ -1065,43 +1348,28 @@ render_svg(struct icon *icon, int x, int y, int size,
     assert(icon->type == ICON_SVG);
 
     if (icon->svg == NULL) {
-        struct timeval start, stop, diff;
-        gettimeofday(&start, NULL);
+        struct timespec *start_load = time_begin();
 
-        if (!icon_from_svg(icon, icon->path))
+        if (!icon_from_svg(icon, icon->path)) {
+            free(start_load);
             return;
-
-        if (print_timing_info) {
-            gettimeofday(&stop, NULL);
-            timersub(&stop, &start, &diff);
-
-            LOG_WARN("%s loaded in %llus %lluµs",
-                     icon->path,
-                     (unsigned long long)diff.tv_sec,
-                     (unsigned long long)diff.tv_usec);
         }
 
+        time_finish(start_load, NULL, "%s loaded", icon->path);
         LOG_DBG("%s", icon->path);
     }
 
-    struct timeval start, stop, diff;
-    gettimeofday(&start, NULL);
+    struct timespec *render_start = time_begin();
 
 #if defined(FUZZEL_ENABLE_SVG_LIBRSVG)
     render_svg_librsvg(icon, x, y, size, cairo);
 #elif defined(FUZZEL_ENABLE_SVG_NANOSVG)
     render_svg_nanosvg(icon, x, y, size, pix, cairo, gamma_correct);
+#elif defined(FUZZEL_ENABLE_SVG_RESVG)
+    render_svg_resvg(icon, x, y, size, pix, cairo, gamma_correct);
 #endif
 
-    if (print_timing_info) {
-        gettimeofday(&stop, NULL);
-        timersub(&stop, &start, &diff);
-
-        LOG_WARN("%s rendered in %llus %lluµs",
-                 icon->path,
-                 (unsigned long long)diff.tv_sec,
-                 (unsigned long long)diff.tv_usec);
-    }
+    time_finish(render_start, NULL, "%s rendered", icon->path);
 }
 
 #if defined(FUZZEL_ENABLE_PNG_LIBPNG)
@@ -1274,39 +1542,24 @@ render_png(struct icon *icon, int x, int y, int size, pixman_image_t *pix,
     assert(icon->type == ICON_PNG);
 
     if (icon->png == NULL) {
-        struct timeval start, stop, diff;
-        gettimeofday(&start, NULL);
+        struct timespec *start_load = time_begin();
 
-        if (!icon_from_png(icon, icon->path, gamma_correct))
+        if (!icon_from_png(icon, icon->path, gamma_correct)) {
+            free(start_load);
             return;
-
-        if (print_timing_info) {
-            gettimeofday(&stop, NULL);
-            timersub(&stop, &start, &diff);
-            LOG_WARN("%s loaded in %llus %lluµs",
-                     icon->path,
-                     (unsigned long long)diff.tv_sec,
-                     (unsigned long long)diff.tv_usec);
         }
 
+        time_finish(start_load, NULL, "%s loaded", icon->path);
         LOG_DBG("%s", icon->path);
     }
 
-    struct timeval start, stop, diff;
-    gettimeofday(&start, NULL);
+    struct timespec *start_render = time_begin();
 
 #if defined(FUZZEL_ENABLE_PNG_LIBPNG)
     render_png_libpng(icon, x, y, size, pix, cairo, scaling_filter);
 #endif
 
-    if (print_timing_info) {
-        gettimeofday(&stop, NULL);
-        timersub(&stop, &start, &diff);
-        LOG_WARN("%s rendered in %llus %lluµs",
-                 icon->path,
-                 (unsigned long long)diff.tv_sec,
-                 (unsigned long long)diff.tv_usec);
-    }
+    time_finish(start_render, NULL, "%s rendered", icon->path);
 }
 
 static int
@@ -1315,7 +1568,8 @@ first_row_y(const struct render *render)
     return (render->border_size +
             render->y_margin +
             (render->conf->hide_prompt ? 0 : render->row_height) +
-            (render->conf->hide_prompt ? 0 : render->inner_pad));
+            (render->conf->hide_prompt ? 0 : render->inner_pad) +
+            render->message_height);
 }
 
 static void
@@ -1692,6 +1946,7 @@ render_initialize_colors(struct render *render, const struct config *conf,
     render->pix_background_color = rgba2pixman(gamma_correct, conf->colors.background);
     render->pix_border_color = rgba2pixman(gamma_correct, conf->colors.border);
     render->pix_text_color = rgba2pixman(gamma_correct, conf->colors.text);
+    render->pix_message_color = rgba2pixman(gamma_correct, conf->colors.message);
     render->pix_prompt_color = rgba2pixman(gamma_correct, conf->colors.prompt);
     render->pix_input_color = rgba2pixman(gamma_correct, conf->colors.input);
     render->pix_match_color = rgba2pixman(gamma_correct, conf->colors.match);
@@ -1736,12 +1991,33 @@ render_resized(struct render *render, int *new_width, int *new_height)
 
     const unsigned icon_height = max(0, row_height - font->descent);
 
+    unsigned message_lines = 0;
+    unsigned longest_message_line = 0;
+    if (render->conf->message != NULL) {
+        char32_t* current_pos = render->conf->message;
+        char32_t* current_line_end = NULL;
+        char32_t* next_line_start = NULL;
+        while (current_pos != NULL) {
+            message_lines++;
+            current_line_boundaries(current_pos, render->conf,
+                                    &current_line_end, &next_line_start);
+
+            if (render->conf->message_mode == MESSAGE_MODE_EXPAND
+                && current_line_end + 1 - current_pos > longest_message_line)
+            {
+                longest_message_line = current_line_end + 1 - current_pos;
+            }
+            current_pos = next_line_start;
+        }
+    }
+
     const unsigned height =
         border_size +                        /* Top border */
         y_margin +
         (render->conf->hide_prompt ? 0 : row_height) +          /* The prompt (hidden if hide_prompt) */
         (render->conf->hide_prompt ? 0 : inner_pad) +           /* Padding between prompt and matches (only if prompt shown) */
         render->conf->lines * row_height +   /* Matches */
+        message_lines * row_height +         /* Message */
         y_margin +
         border_size;                         /* Bottom border */
 
@@ -1751,7 +2027,7 @@ render_resized(struct render *render, int *new_width, int *new_height)
         (max((W->advance.x + pt_or_px_as_pixels(
                   render, &render->conf->letter_spacing)),
              0)
-         * render->conf->chars) +
+         * max(render->conf->chars, longest_message_line)) +
         x_margin +
         border_size;
 
@@ -1768,6 +2044,7 @@ render_resized(struct render *render, int *new_width, int *new_height)
     render->selection_border_radius = selection_border_radius;
     render->row_height = row_height;
     render->icon_height = icon_height;
+    render->message_height = message_lines * render->row_height;
 
     if (new_width != NULL)
         *new_width = width;
@@ -1848,6 +2125,7 @@ render_destroy(struct render *render)
     tll_free(render->workers.queue);
 
     fcft_text_run_destroy(render->prompt_text_run);
+    fcft_text_run_destroy(render->message_text_run);
     fcft_text_run_destroy(render->placeholder_text_run);
 
     if (render->selection_corners != NULL)
@@ -1866,11 +2144,12 @@ render_get_row_num(const struct render *render, int width, int x, int y,
     const int inner_pad = render->inner_pad;
     const int border_size = render->border_size;
     const int row_height = render->row_height;
+    const int message_height = render->message_height;
 
     const int min_x = render->border_size + render->x_margin - render->x_margin / 3;
     const int max_x = width - (min_x);
 
-    const int first_row = 1 * border_size + y_margin +
+    const int first_row = message_height + border_size + y_margin +
         (render->conf->hide_prompt ? 0 : row_height) +
         (render->conf->hide_prompt ? 0 : inner_pad);
 
@@ -1889,7 +2168,9 @@ void
 render_flush_text_run_cache(struct render *render)
 {
     fcft_text_run_destroy(render->prompt_text_run);
+    fcft_text_run_destroy(render->message_text_run);
     fcft_text_run_destroy(render->placeholder_text_run);
     render->prompt_text_run = NULL;
+    render->message_text_run = NULL;
     render->placeholder_text_run = NULL;
 }
